"""Assembly."""
from collections import namedtuple
from copy import deepcopy
from itertools import groupby
from itertools import zip_longest
from typing import Any
from typing import Dict
from typing import Generator
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Union

import networkx as nx
import numpy as np
import pandas as pd
import primer3plus
from Bio.Alphabet import generic_dna
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from more_itertools import pairwise
from primer3plus.utils import reverse_complement as rc
from pyblast.utils import is_circular

from .alignment import Alignment
from .alignment import AlignmentGroup
from .alignment import PCRProductAlignmentGroup
from .alignment_container import AlignmentContainer
from .molecule import Molecule
from .molecule import MoleculeType
from .molecule import Reaction
from dasi.constants import Constants
from dasi.cost import SpanCost
from dasi.exceptions import DasiDesignException
from dasi.exceptions import DasiInvalidMolecularAssembly
from dasi.exceptions import DasiNoPrimerPairsException
from dasi.exceptions import DasiSequenceDesignException
from dasi.log import logger
from dasi.utils import NumpyDataFrame
from dasi.utils import Region
from dasi.utils import sort_cycle


def _design_primers(
    template: str,
    region: Region,
    lseq: Union[None, str],
    rseq: Union[None, str],
    left_overhang: Union[None, str] = None,
    right_overhang: Union[None, str] = None,
) -> Tuple[Dict[int, dict], Dict[str, Any]]:
    """Design primers flanking the specified.

    :class:`Region.<dasi.utils.Region>`. If the region is cyclic and spans the
    origin, this method will handle the appropriate manipulations to design
    primers around the origin and restore the locations of the resulting primer
    pairs.

    :param template: the template string to design primers
    :param region: region specified to design primers around. Regions are exclusive at
                    their end points (`.b` parameter)
    :param lseq: optionally provided left sequence
    :param rseq: optionally provided right sequence
    :param left_overhang: optionally provided left overhang sequence of the primer
    :param right_overhang: optionally provided right overhang sequence of the primer
    :return: tuple of pairs and the 'explain' dictionary.
    """
    design = primer3plus.new()
    design.settings.as_cloning_task()
    if region.direction == -1:
        region = region.flip()
        template = rc(template)

    if lseq and left_overhang:
        raise DasiSequenceDesignException
    if rseq and right_overhang:
        raise DasiSequenceDesignException

    if region.spans_origin():
        adjusted_template = region.get_slice(template) + region.invert()[0].get_slice(
            template
        )
        design.settings.template(adjusted_template)
        design.settings.included((0, len(region)))
        index = list(region) + list(region.invert()[0])
    else:
        design.settings.template(template)
        design.settings.included((region.a, len(region)))
        index = None
    if lseq:
        design.settings.left_sequence(lseq)
    if rseq:
        design.settings.right_sequence(rseq)

    if left_overhang is None:
        left_overhang = ""
    if right_overhang is None:
        right_overhang = ""

    design.settings.product_size((len(region), len(region)))
    design.settings.left_overhang(left_overhang)
    design.settings.right_overhang(right_overhang)
    design.PRIMER_PICK_ANYWAY = True
    design.settings.use_overhangs()
    design.settings.long_ok()

    design.logger.set_level("INFO")
    pairs, explain = design.run_and_optimize(15)
    if index is not None:
        for pair in pairs.values():
            loc = pair["LEFT"]["location"]
            pair["LEFT"]["location"] = (index[loc[0]], loc[1])

            loc = pair["RIGHT"]["location"]
            pair["RIGHT"]["location"] = (index[loc[0]], loc[1])
    return pairs, explain


def _edata_to_npdf(edata: dict, span_cost: SpanCost) -> NumpyDataFrame:
    return span_cost.cost(edata["span"], edata["type_def"])


def _no_none_or_nan(*i):
    for _i in i:
        if _i is not None and not np.isnan(_i):
            return _i


def _get_primer_extensions(
    graph: nx.DiGraph, n1: "AssemblyNode", n2: "AssemblyNode", cyclic: bool = True
) -> Tuple[int, int]:
    """Return the left and right primer extensions for the given *internal*
    fragment. To get the extensions, we look for the left predecessor edge and
    get its `right_ext` or `rprimer_right_ext` and on the other side the right
    successor edge and its `left_ext` or `lprimer_left_ext`.

    :param graph: assembly graph
    :param n1: source node on the graph
    :param n2: end node on the graph
    :param cyclic: whether the
    :return: tuple of left and right extensions
    """
    # NOTE: the successors 'left' primer is the fragments 'right' primer extension
    successors = list(graph.successors(n2))
    if successors:
        sedge = graph[n2][successors[0]]
        right_ext = _no_none_or_nan(sedge["lprimer_left_ext"], sedge["left_ext"])
    elif cyclic:
        raise DasiSequenceDesignException(
            "Sequence is cyclic but there are no " "successors for {}".format(n2)
        )
    else:
        right_ext = 0

    # the predecessors 'right' primer is the fragments 'left' primer extension
    predecessors = list(graph.predecessors(n1))
    if predecessors:
        pedge = graph[predecessors[0]][n1]
        left_ext = _no_none_or_nan(pedge["rprimer_right_ext"], pedge["right_ext"])
    elif cyclic:
        raise DasiSequenceDesignException(
            "Sequence is cyclic but there are no " "precessors for {}".format(n1)
        )
    else:
        left_ext = 0
    return int(left_ext), int(right_ext)


def _use_direct(
    edge: Tuple["AssemblyNode", "AssemblyNode", dict], seqdb: Dict[str, SeqRecord]
) -> Tuple[SeqRecord, AlignmentGroup]:
    group = edge[2]["groups"][0]
    alignment = group.alignments[0]
    sk = alignment.subject_keys
    srecord = seqdb[sk]
    return srecord, alignment


def _design_gap(
    edge: Tuple["AssemblyNode", "AssemblyNode", dict], qrecord: SeqRecord
) -> Union[Tuple[SeqRecord, Region], Tuple[None, None]]:
    n1, _, edge_data = edge
    gene_size = edge_data["gene_size"]
    if not np.isnan(gene_size):
        lshift = edge_data["lshift"]
        assert not np.isnan(lshift)
        a = n1.index + lshift
        b = a + gene_size
        gene_region = edge_data["query_region"].new(a, b)
        gene_seq = gene_region.get_slice(qrecord)
        return gene_seq, gene_region
    else:
        return None, None


def _design_pcr_product_primers(
    edge: Tuple["AssemblyNode", "AssemblyNode", dict],
    graph: nx.DiGraph,
    design: Tuple[bool, bool],
    seqdb: Dict[str, SeqRecord],
) -> Tuple[Dict, Dict, AlignmentGroup, Region]:
    if edge[-1]["type_def"].int_or_ext == "external":
        raise Exception()

    # this is a new PCR product
    n1, n2, edata = edge
    lext, rext = _get_primer_extensions(graph, n1, n2)
    alignment_groups = edata["groups"]
    group = alignment_groups[0]

    qkey = group.query_key
    qrecord = seqdb[qkey]
    qregion = group.query_region

    # set overhangs
    loverhang = ""
    roverhang = ""
    if design[0]:
        lregion = qregion.new(qregion.a - lext, qregion.a)
        loverhang = lregion.get_slice(qrecord.seq, as_type=str)
    if design[1]:
        rregion = qregion.new(qregion.b, qregion.b + rext)
        roverhang = primer3plus.utils.reverse_complement(
            rregion.get_slice(qrecord.seq, as_type=str)
        )

    # TODO: complex alignment groups have re-adjusted subject regions
    # TODO: ensure fwd and rev primers have same template
    # TODO: the 'fwd' primer of a template in reverse direction is its reverse primer,
    #       so get_template needs to be adjusted.
    # collect template, left primer, and right primer keys

    lkey, rkey = None, None
    fwd, template, rev = None, None, None
    if design == (1, 1):
        assert isinstance(group, AlignmentGroup)
        tkey = group.subject_keys[0]
        template = group.alignments[0]
    else:
        if not hasattr(group, "groupings"):
            print(group)
            print(group.type)
        grouping = group.groupings[0]
        template = group.get_template(0)
        assert grouping["template"].subject_key == template.subject_key

        # get primer keys
        fwd = grouping["fwd"]
        rev = grouping["rev"]
        tkey = template.subject_key
        if fwd:
            lkey = fwd.subject_key
        if rev:
            rkey = rev.subject_key

    if not design[1]:
        roverhang = ""

    if not design[0]:
        loverhang = ""

    trecord = seqdb[tkey]
    tseq = str(trecord.seq)
    if rkey:
        rrecord = seqdb[rkey]
        rseq = str(rrecord.seq)
    else:
        rseq = None
    if lkey:
        lrecord = seqdb[lkey]
        lseq = str(lrecord.seq)
    else:
        lseq = None

    # design primers
    pairs, explain = _design_primers(
        tseq,
        template.subject_region,
        lseq,
        rseq,
        left_overhang=loverhang,
        right_overhang=roverhang,
    )
    for pair in pairs.values():
        pair["LEFT"]["SUBJECT_KEY"] = lkey
        pair["LEFT"]["GROUP"] = fwd
        pair["RIGHT"]["SUBJECT_KEY"] = rkey
        pair["RIGHT"]["GROUP"] = rev
        pair["PAIR"]["SUBJECT_KEY"] = tkey
        pair["PAIR"]["GROUP"] = template

    query_region = group.query_region.new(
        group.query_region.a - len(loverhang) + group.query_region.context_length,
        group.query_region.c + len(roverhang) + group.query_region.context_length,
    )
    #     print(group.query_region)
    #     print(template.query_region)
    #     print(group)
    #     print(len(query_region), len(group.query_region) + lext + rext, lext, rext)
    #     assert len(query_region) == len(group.query_region) + lext + rext
    return pairs, explain, template, query_region


def _design_edge(
    assembly: "Assembly",
    n1: "AssemblyNode",
    n2: "AssemblyNode",
    seqdb: Dict[str, SeqRecord],
) -> Union[Reaction, None]:
    query_key = assembly.query_key
    graph = assembly.graph

    edge = n1, n2, graph[n1][n2]

    moltype = edge[2]["type_def"]
    qrecord = seqdb[query_key]
    # contains information about templates and queries

    if edge[-1]["type_def"].int_or_ext == "external":
        if moltype.use_direct:
            # this is a fragment used directly in an assembly
            frag_seq, frag_alignment = _use_direct(edge, seqdb)
            frag_mol = Molecule(moltype, frag_alignment, frag_seq)
            return Reaction(
                Reaction.Types.Direct, inputs=[], outputs=[frag_mol], metadata=edge[2]
            )
        elif moltype.synthesize:
            # this is either a gene synthesis fragment or already covered by the primers.
            synthesis_seq, synthesis_region = _design_gap(edge, qrecord)
            if synthesis_seq:
                subject_region = Region(
                    0, len(synthesis_region), len(synthesis_region), direction=1
                )
                synthesis_alignment = Alignment(
                    synthesis_region, subject_region, moltype.name, query_key, ""
                )
                synthesis_mol = Molecule(
                    moltype,
                    synthesis_alignment,
                    synthesis_seq,
                    query_region=synthesis_region,
                )
                return Reaction(
                    Reaction.Types.Synthesize,
                    inputs=[],
                    outputs=[synthesis_mol],
                    metadata=edge[2],
                )
            else:
                return None
        else:
            return None
    elif edge[-1]["type_def"].name == Constants.FRAGMENT:
        group = edge[2]["groups"][0]
        alignment = group.alignments[0]
        sk = alignment.subject_key
        frag_seq = seqdb[sk]
        query_region = edge[2]["query_region"]
        frag_mol = Molecule(moltype, alignment, frag_seq, query_region=query_region)
        return Reaction(
            Reaction.Types.Direct, inputs=[], outputs=[frag_mol], metadata=edge[2]
        )
    elif edge[-1]["type_def"].name == Constants.SHARED_SYNTHESIZED_FRAGMENT:
        query_region = edge[2]["query_region"]
        group = edge[2]["group"]
        synthesis_mol = Molecule(
            MoleculeType.types[Constants.SHARED_SYNTHESIZED_FRAGMENT],
            alignment_group=group,
            sequence=query_region.get_slice(seqdb[query_key]),
            query_region=query_region,
        )
        return Reaction(
            Reaction.Types.Synthesize,
            inputs=[],
            outputs=[synthesis_mol],
            metadata=edge[2],
        )
    else:
        pairs, explain, group, query_region = _design_pcr_product_primers(
            edge, graph, moltype.design, seqdb
        )
        if not pairs:
            raise DasiNoPrimerPairsException("No primer pairs were found.")
        pair = pairs[0]
        primers = []
        for x in ["LEFT", "RIGHT"]:
            primer_seq = pair[x]["OVERHANG"] + pair[x]["SEQUENCE"]
            primer_group = pair[x]["GROUP"]
            primer_record = SeqRecord(Seq(primer_seq, generic_dna))
            primer = Molecule(
                MoleculeType.types[Constants.PRIMER],
                primer_group,
                primer_record,
                metadata=pair[x],
            )
            primers.append(primer)
        template = Molecule(
            MoleculeType.types[Constants.TEMPLATE],
            pair["PAIR"]["GROUP"],
            seqdb[pair["PAIR"]["SUBJECT_KEY"]],
        )

        product = Molecule(
            moltype,
            group,
            query_region.get_slice(seqdb[query_key]),
            query_region=query_region,
        )

        return Reaction(
            Reaction.Types.PCR,
            inputs=primers + [template],
            outputs=[product],
            metadata=edge[2],
        )


AssemblyNode = namedtuple(
    "AssemblyNode", "index expandable type overhang"
)  #: tuple representing a location on a goal sequence


class Assembly(Iterable):
    """Should take in a path, graph, container, seqdb to produce relevant
    information."""

    REACTION_DF_COLS = (
        "DESIGN_ID",
        "DESIGN_KEY",
        "ASSEMBLY_ID",
        "REACTION_ID",
        "REACTION_NAME",
        "NAME",
        "TYPE",
        "KEY",
        "ROLE",
        "REGION",
        "SEQUENCE",
        "LENGTH",
        "META",
    )
    REACTION_DF_SORT_BY = (
        "TYPE",
        "DESIGN_ID",
        "ASSEMBLY_ID",
        "REACTION_ID",
        "REACTION_NAME",
        "NAME",
        "ROLE",
    )
    SUMMARY_DF_COLS = (
        "query_start",
        "query_end",
        "subject_names",
        "subject_keys",
        "subject_starts",
        "subject_ends",
        "cost",
        "material",
        "span",
        "type",
        "internal_or_external",
        "efficiency",
        "complexity",
        "notes",
    )

    def __init__(
        self,
        nodes: List[AssemblyNode],
        container: AlignmentContainer,
        full_assembly_graph: nx.DiGraph,
        query_key: str,
        query: SeqRecord,
        seqdb,
        do_raise: bool = True,
    ):
        self.logger = logger(self)
        self._nodes = tuple(nodes)
        self._reactions = tuple()
        self.validate_input_nodes()
        self.seqdb = seqdb
        self.query_key = query_key
        self.query = query
        self._full_graph = full_assembly_graph

        self.container = container
        self.groups = container.groups()
        if len(self.groups) == 0:
            raise DasiDesignException("No groups were found in container.")
        self.graph = self._subgraph(self._full_graph, nodes, do_raise=do_raise)
        nx.freeze(self.graph)

        if do_raise:
            self.post_validate(do_raise)

    @property
    def reactions(self):

        top_level_reaction = Reaction(
            Reaction.Types.Assembly,
            inputs=[],
            outputs=[],
            metadata={"query_key": self.query_key},
        )
        top_level_reaction.outputs = [
            Molecule(
                molecule_type=MoleculeType.types[Constants.PLASMID],
                sequence=deepcopy(self.query),
                alignment_group=None,
            )
        ]

        if not self._reactions:
            reactions = [top_level_reaction]
            for n1, n2, edata in self.edges():
                reaction = _design_edge(self, n1, n2, seqdb=self.seqdb)
                if reaction:
                    for _mol in reaction.outputs:
                        top_level_reaction.inputs.append(_mol)
                    reactions.append(reaction)
            self._reactions = tuple(reactions)
        return self._reactions

    @property
    def assembly_reactions(self):
        return [r for r in self.reactions if r.name == Reaction.Types.Assembly]

    @property
    def nonassembly_reactions(self):
        return [r for r in self.reactions if r.name != Reaction.Types.Assembly]

    def post_validate(self):
        total_span = 0
        for n1, n2, edata in self.edges():
            if n1.type == n2.type:
                raise DasiInvalidMolecularAssembly("Invalid assembly graph")
            total_span += edata["span"]
        if not total_span == len(self.query):
            raise DasiInvalidMolecularAssembly(
                "Assembly length '{}' is different from expected"
                " length '{}'".format(total_span, len(self.query))
            )

    def _head(self):
        """Get the 'first' 'A' node."""
        x = sorted(list(self.graph.nodes), key=lambda n: (n.type == "B", n.index))
        return x[0]

    def _sorted_edges(self):
        head = self._head()
        edges = list(nx.bfs_edges(self.graph, head))
        edges += [
            (t[1], t[0])
            for t in nx.bfs_edges(self.graph, head, reverse=True, depth_limit=1)
        ]

        return edges

    def validate_input_nodes(self):
        # rule 1 A -> B -> A -> B -> ...
        types = [n.type for n in self._nodes]
        groups = [list(g) for _, g in groupby(types)]
        if len(types) != len(groups):
            raise ValueError("There invalid edges input nodes")

    @staticmethod
    def _missing_edata():
        return {
            "cost": np.inf,
            "weight": np.inf,
            "material": np.inf,
            "efficiency": 0.0,
            "type_def": MoleculeType.types[Constants.MISSING],
            "span": np.inf,
        }

    def _subgraph(
        self, graph: nx.DiGraph, nodes: List[AssemblyNode], do_raise: bool = True
    ):
        def _resolve(node: "AssemblyNode", qregion) -> Tuple[AssemblyNode, dict]:
            new_node = AssemblyNode(qregion.t(node.index), *list(node)[1:])
            return new_node

        subgraph = nx.OrderedDiGraph()
        nodes = [AssemblyNode(*n) for n in nodes]
        example_query_region = self.container.alignments[0].query_region

        resolved_nodes = [_resolve(node, example_query_region) for node in nodes]
        if self.cyclic:
            resolved_nodes = sort_cycle(
                resolved_nodes, key=lambda n: (n.type, n.index, n)
            )
        subgraph.add_nodes_from(resolved_nodes)

        pair_iter = list(pairwise(nodes))

        if self.cyclic:
            pair_iter.append((nodes[-1], nodes[0]))

        visited = set()
        for n1, n2 in pair_iter:
            edata = graph.get_edge_data(n1, n2)
            if edata is None:
                # if n1.index > len(self.query):
                #     n3 = AssemblyNode(n1.index - len(self.query), *list(n1)[1:])
                #     edata = graph.get_edge_data(n3, n2)
                if edata is None:
                    edata = self._missing_edata()
            else:
                assert edata["type_def"].int_or_ext

            query_region = self.container.alignments[0].query_region.new(
                n1.index, n2.index
            )
            groups = self.container.find_groups_by_pos(
                query_region.a,
                query_region.b,
                group_type=edata["type_def"].name,
                groups=self.groups,
            )
            if edata["type_def"].int_or_ext == "internal" and not groups:
                raise DasiDesignException(
                    "Missing groups for edge between {} and {}".format(n1, n2)
                )

            edata["groups"] = groups
            edata["query_region"] = query_region

            rn1 = _resolve(n1, query_region)
            rn2 = _resolve(n2, query_region)

            if rn1 in visited:
                break
            visited.add(rn1)

            # TODO: add this check
            if do_raise:
                if rn1 in subgraph:
                    raise DasiInvalidMolecularAssembly(
                        "Node already exists in subgraph"
                    )
                if rn2 in subgraph:
                    raise DasiInvalidMolecularAssembly(
                        "Node already exists in subgraph"
                    )

            subgraph.add_edge(rn1, rn2, **edata)
        return subgraph

    @property
    def cyclic(self):
        return is_circular(self.query)

    # TODO: consolidate this with shortest path utils in networkx
    def cost(self):
        material = 0
        efficiency = 1.0
        for _, _, edata in self.edges():
            material += edata["material"]
            efficiency *= edata["efficiency"]
        if efficiency == 0:
            return np.inf
        return material / efficiency

    def edges(self, data=True) -> Iterable[Tuple[AssemblyNode, AssemblyNode, Dict]]:
        for n1, n2 in self._sorted_edges():
            if data:
                edata = self.graph[n1][n2]
                yield n1, n2, edata
            else:
                yield n1, n2

    def nodes(self, data=True) -> Iterable[Tuple[AssemblyNode, Dict]]:
        return self.graph.nodes(data=data)

    def edit_distance(
        self, other: "Assembly", explain=False
    ) -> Union[int, Tuple[int, List[Tuple[int, str]]]]:
        differences = []
        for i, (n1, n2) in enumerate(
            zip_longest(self.nodes(data=False), other.nodes(data=False))
        ):
            if n1 is None or n2 is None:
                differences.append((i, "{} != {}".format(n1, n2)))
                continue
            if n1.index != n2.index:
                differences.append((i, "Index: {} != {}".format(n1.index, n2.index)))
            if n1.expandable != n2.expandable:
                differences.append(
                    (i, "Expandable: {} != {}".format(n1.expandable, n2.expandable))
                )
            if n1.type != n2.type:
                differences.append((i, "Type: {} != {}".format(n1.type, n2.type)))
            if n1.overhang != n2.overhang:
                differences.append(
                    (i, "Overhang: {} != {}".format(n1.overhang, n2.overhang))
                )
        dist = len(differences)
        if explain:
            return dist, differences
        return dist

    def print(self):
        print("query_name: {}".format(self.query.name))
        print("query_key: {}".format(self.query_key))
        print("Cost: {}".format(self.cost()))
        df = self.to_df()
        print(df)

    def print_diff(self, other: "Assembly"):
        for i, (n1, n2) in enumerate(
            zip_longest(self.nodes(data=False), other.nodes(data=False))
        ):
            if n1 != n2:
                desc = False
            else:
                desc = True
            print("{} {} {}".format(desc, n1, n2))

    def to_df(self) -> pd.DataFrame:
        rows = []
        for n1, n2, edata in self.edges():
            groups = edata["groups"]

            if groups:
                group = groups[0]
                if isinstance(group, PCRProductAlignmentGroup):
                    alignments = group.alignments
                else:
                    alignments = group.alignments[:1]
            else:
                alignments = []
            subject_keys = [a.subject_key for a in alignments]
            subject_names = [self.container.seqdb[key].name for key in subject_keys]
            subject_starts = [a.subject_region.a for a in alignments]
            subject_ends = [a.subject_region.b for a in alignments]

            rows.append(
                {
                    "query_start": edata["query_region"].a,
                    "query_end": edata["query_region"].b,
                    "subject_names": subject_names,
                    "subject_keys": subject_keys,
                    "subject_starts": subject_starts,
                    "subject_ends": subject_ends,
                    "cost": edata["cost"],
                    "material": edata["material"],
                    "span": edata["span"],
                    "type": edata["type_def"].name,
                    "internal_or_external": edata["type_def"].int_or_ext,
                    "efficiency": edata.get("efficiency", np.nan),
                    "complexity": edata.get("complexity", np.nan),
                    "notes": edata.get("notes", ""),
                }
            )

        df = pd.DataFrame(rows)
        assert set(df.columns) == set(self.SUMMARY_DF_COLS)
        return df

    def _csv_row(
        self,
        m: Molecule,
        role: str,
        reaction_id: Union[str, int],
        reaction_name: str,
        meta: dict = None,
    ) -> dict:
        mtype = m.type.name
        group = m.alignment_group

        key = None
        name = None
        if group:
            if hasattr(group, "subject_key"):
                if group.subject_key:
                    key = group.subject_key
                    name = self.seqdb[key].name
            else:
                # TODO: select best subject here
                key = group.subject_keys[0]
                name = self.seqdb[key].name

        if m.query_region:
            q = (m.query_region.a, m.query_region.b, m.query_region.context_length)
        else:
            q = None
        length = len(m.sequence)
        data = {
            "NAME": name,
            "LENGTH": length,
            "SEQUENCE": str(m.sequence.seq),
            "TYPE": mtype,
            "KEY": key,
            "REGION": q,
            "ROLE": role,
            "REACTION_ID": reaction_id,
            "REACTION_NAME": reaction_name,
        }
        if meta:
            data.update({"META": deepcopy(meta)})
        return data

    @property
    def molecules(self) -> Generator[Tuple[int, str, Molecule], None, None]:
        for i, r in enumerate(self.reactions):
            for m in r.inputs:
                yield (i, "input", m)
            for m in r.outputs:
                yield (i, "output", m)

    def to_reaction_df(self) -> pd.DataFrame:
        rows = []
        reactions = self.reactions
        for reaction_id, role, m in self.molecules:
            reaction = reactions[reaction_id]
            if m.type.name == "PRIMER":
                meta = deepcopy(m.metadata)
                meta["ANNEAL"] = meta["SEQUENCE"]
                del meta["SEQUENCE"]
                meta = {"PRIMER_{}".format(k): v for k, v in meta.items()}
            else:
                meta = None
            rows.append(self._csv_row(m, role, reaction_id, reaction.name, meta))
        colnames = self.REACTION_DF_COLS
        df = pd.DataFrame(rows, columns=colnames)
        df.sort_values(by=list(self.REACTION_DF_SORT_BY), inplace=True)
        return df

    def __eq__(self, other: "Assembly") -> bool:
        return self.edit_distance(other) == 0

    def __iter__(self) -> Generator[AssemblyNode, None, None]:
        for n in self.nodes(data=False):
            yield n



r"""
Design (:mod:`dasi.design`)
=============================

.. currentmodule:: dasi.design

This module provides representations of Molecules and Molecular Assemblies.

.. autosummary::
    :toctree: generated/

    Design
    LibraryDesign
    DesignResult
"""
from .design import Design
from .design import DesignResult
from .library_design import LibraryDesign



import operator
from collections.abc import Container
from collections.abc import Sized
from functools import reduce
from itertools import chain
from typing import Any
from typing import Callable
from typing import Generator
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Union


class SpanError(Exception):
    pass


# TODO: move this to pyblast?
class Span(Container, Iterable, Sized):
    """`Span` maps the provided positions onto a context.

    Spans have no direction and have an underlying context
    that has a certain length, can be linear or cyclic, and has a starting index.

    **Linear spans**
    For example, a basic linear span can be represented using the following:

    .. code-block:: python

        s = Span(0, 10, 20)
        assert s.a == 0
        assert s.b == 10
        assert list(s) == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]  # spans are always
                                                          # exclusive at endpoint.

    **Indexing and Slicing**

    New indexes can be provided to spans.

    .. code-block:: python

        s = Span(1, 10, 10, index=1)
        assert s.a == 1
        assert s.b == 10

    Positions can be mapped to the span:

    .. code-block:: python

        s = Span(5, 8, 10, index=5)
        assert s.a == 5
        assert s.b == 8
        assert s[0] == 5    # the 'first' position in the span equivalent to the
                            # starting index
        assert s[-1] == 7   # the 'last' position in the span is the last inclusive
                            # index

    Positions can be mapped automatically during initialization:

    .. code-block:: python

        s = Span(-1, 5, 10, cyclic=True, index=0)   # index '-1' is mapped onto last
                                                    # available index on the context,
        '9'
        assert s.a == 9
        assert s.b == 5

        s = Span(6, -1, 10, cyclic=True, index=0)  # index '-1' is mapped onto last
                                                   # available exclusive index on the
         context, '10'
        assert s.a == 6
        assert s.b == 10

    Inclusive positions can be mapped onto the context using `t`:

    .. code-block:: python

        s = Span(-1, 5, 10, cyclic=True)
        assert s.t(11) == 1
        assert s.t(10) == 0

    **Reindexing**

    Context starting index can be remapped:

    .. code-block:: python

        s = Span(0, 5, 10)
        s1 = s.reindex(1)
        asset s1.a == 1
        assert s2.b == 6

    **Slicing**

    Spans can be sliced similarly to lists:

    .. code-block:: python

        s = Span(4, 8, 10)
        s1 = s[1:]
        assert s1.a == 5
        assert s2.b == 8

        s2 = s[:-1]
        assert s2.a == 4
        assert s2.b == 7

        s3 = s[2:-1]
        assert s3.a == 5
        assert s3.b == 7

    **Cyclic spans**
    A cyclic span can be represented in the following way:

    .. code-block:: python

        s = Span(18, 2, 20, cyclic=True)
        assert s.a == 18
        assert s.b == 2
        assert list(s) == [18, 19, 0, 1]
        assert len(s) == 4

    **Wrapping cyclic spans**
    Spans the wrap around the context multiple times can be represented as well.
    The mapped endpoint
    is found with `span.b` and the non-mapped endpoint is found using `span.c`.

    .. code-block:: python

        s = Span(9, 21, 10, cyclic=True)
        assert list(s) == [9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0]
        assert len(s) == 12
        assert s.a == 9  # start point
        assert s.b == 1  # mapped endpoint
        assert s.c == 21 # the unmapped endpoint

    The indices are reduced to their lowest 'wrapping' whenever possible.
    For example, the following
    initializations are equivalent:

    .. code-block:: python

        s1 = Span(1, 5, 10, cyclic=True)
        s2 = Span(11, 15, 10, cyclic=True)
        s3 = Span(21, 25, 10, cyclic=True)
        assert s1 == s2
        assert s2 == s3

    **Iterating through spans**

    Spans can be treated like iterators:

    .. code-block:: python

        for i in Span(5, 3, 10, cyclic=True):
            print(i)

    **Span inversion**

    Inversion returns two spans that represent everything *except* the span:

    .. code-block:: python

        s = Span(4, 8, 10)
        s1, s2 = s.invert()  # also s[::-1]
        assert s1.a == 0
        assert s1.b == 4
        assert s2.a == 8
        assert s2.b == 10

        # inverted cyclic span returns one valid span and None
        s = Span(4, 8, 10, cyclic=True)
        s1, s2 = s.invert()
        assert s1.a == 8
        assert s1.b == 4
        assert s2 is None

    **Intersections, differences**

    .. code-block:: python

        s1 = Span(4, 10, 20)
        s2 = Span(8, 12, 20)
        s3 = s1.intersection(s2)
        assert s3.a == 8
        assert s3.b == 10

    .. code-block:: python

        s1 = Span(4, 10, 20)
        s2 = Span(8, 12, 20)
        s3, s4 = s1.differences(s2)

        assert s3.a, s3.b == 4, 8
        assert s4,a, s4.b == 10, 12

    **Gotchas**

    .. code-block:: python

        s2 = Span(8, 2, 10, cyclic=True)
        assert s2.ranges() == [(8, 10), (0, 2)]
        assert s2.ranges(ignore_wraps=True) == [(8, 10), (0, 2)]

        # # TODO: this is a strange behavior?
        s2 = Span(8, 12, 10, cyclic=True)
        assert s2.ranges() == [(8, 10), (0, 2)]
        assert s2.ranges(ignore_wraps=True) == [(8, 10), (0, 2)]
    """

    __slots__ = [
        "_a",
        "_b",
        "_c",
        "_context_length",
        "_cyclic",
        "_index",
        "_strict",
        "_abs_wrap",
        "_ignore_wrap",
    ]

    def __init__(
        self,
        a: int,
        b: int,
        l: int,
        cyclic=False,
        index=0,
        ignore_wrap=False,
        strict=False,
        abs_wrap=False,
    ):
        """Constructs a new Span. There are several options to customize the
        initialization procedure.

        **strict=True**

        When strict, any index outside the valid bounds of the context raises an
        IndexError.

        .. code-block:: python

            Span(1, 10, 10, cyclic=True, strict=True) # no raise
            Span(1, 11, 10, cyclic=True, strict=True) # raises IndexError
            Span(0, 11, 10, index=1, cyclic=True, strict=True) # raises IndexError

        **ignore_wrap=True**

        When wrapping is ignored, indices are simply mapped to the context with no
        consideration
        of the number of times the absolute position would wrap around the context.

        .. code-block:: python

            # all of the following are equivalent with ignore_wrap == True
            Span(1, 10, 10, cyclic=True, ignore_wrap=True)
            Span(1+10, 10, 10, cyclic=True, ignore_wrap=True)
            Span(1, 10-100, 10, cyclic=True, ignore_wrap=True)
            Span(1+20, 10-100, 10, cyclic=True, ignore_wrap=True)

        **abs_wrap=True**

        When absolute wrapping is used, the absolute difference between starting and
        ending index wrappings
        is calculated, the starting index is to the context while the ending index is
        adjusted such
        that the length will reflect the abs difference between starting and ending
        index wrappings.
        This can be unintuitive
        is best shown with the following example:

        .. code-block:: python

            # all of the following initializations result in equivalent spans

            # starts at 1, wraps around one full time, ends at 5.
            # Length is 15 - 1
            s1 = Span(1, 15, 10, cyclic=True, abs_wrap=True)
            assert len(s1) == 14

            # starts at 11, wraps around one full time, ends at 15.
            # Then positions are mapped back to context at 1 and 5.
            # Length is still 25 - 11 == 14
            s2 = Span(11, 25, 10, cyclic=True, abs_wrap=True)
            assert len(s2) == 14

            # starts at 11, wraps around one full time, ends at 15.
            # Then positions are mapped back to context at 1 and 5.
            # Length is now 11 + 5 = 14
            s3 = Span(11, 5, 10, cyclic=True, abs_wrap=True)
            assert len(s3) == 14

            # the lengths change with the abs diff in number of times wrapped
            _s = Span(21, 5, 10, cyclic=True, abs_wrap=True)
            assert len(_s) == 24

            _s = Span(21, 15, 10, cyclic=True, abs_wrap=True)
            assert len(_s) == 14

            # all
            assert s1 == s2
            assert s2 == s3

        :param a: start of the span (inclusive)
        :type a: int
        :param b: end of the span (exclusive)
        :type b: int
        :param l: context length of the region
        :type l: int
        :param cyclic: whether the underlying context is cyclic
        :type cyclic: bool
        :param index: the starting index of the region
        :type index: int
        :param strict: if True, positions outside of context bounds are disallowed.
        :type strict: bool
        :param ignore_wrap: if True (default False), initialization indicies that wrap
                            around multiple times will
                            simply be mapped directly to the context (no wrapping used).
        :type ignore_wrap: bool
        :param abs_wrap:    if True, the abs difference between start and end wrappings
                            are used. Starting wraps that are greater than ending wraps
                            are valid. If False and the starting wrap is greater than
                            the ending wrap, an IndexError is thrown.
        """
        if isinstance(a, tuple):
            print(a)
        if isinstance(b, tuple):
            print(b)
        a = int(a)
        b = int(b)
        self._context_length = int(l)
        self._index = index
        self._cyclic = cyclic
        self._strict = strict
        self._abs_wrap = abs_wrap
        self._ignore_wrap = ignore_wrap

        # special empty edge case
        if cyclic and a == b:
            self._set_as_empty(a)
            return

        # check bounds
        if self._strict or not cyclic:
            bounds = self.bounds()
            if not bounds[0] <= a < bounds[1]:
                raise IndexError(
                    "Start {} must be in [{}, {})".format(a, index, index + l)
                )
            if not bounds[0] <= b <= bounds[1]:
                raise IndexError(
                    "End {} must be in [{}, {}]".format(b, index, index + l)
                )

        if self._ignore_wrap:
            start_wrap = 0
            end_wrap = 0
        else:
            start_wrap = int((a - index) / l)
            end_wrap = int((b - index - 1) / l)

        if self._abs_wrap is False and start_wrap > end_wrap:
            self._a = a
            self._b = b
            self._c = b
            diff = end_wrap - start_wrap
            # return self._set_as_empty(a)
            raise IndexError(
                "Could not interpret span {span}. Starting position wraps around "
                "context {i} times and end position wraps around {j} times."
                " A valid initialization would be Span({a}, {b}, ...)".format(
                    span=self,
                    i=start_wrap,
                    j=end_wrap,
                    a=self._a,
                    b=self._b - diff * self._context_length,
                )
            )

        # set indices
        _a = a - index
        _b = b - index

        if _a >= l or _a < 0:
            self._a = self.t(_a, False)
        else:
            self._a = a
        if _b > l:
            self._b = self.t(_b - 1, False) + 1
        elif _b < 0:
            self._b = self.t(_b, False) + 1
        else:
            self._b = b

        if self._a > self._b and not cyclic:
            raise IndexError(
                "Start {} cannot be greater than end {} for linear spans.".format(
                    self._a, self._b
                )
            )

        # allow wrap mean this will keep track of how many time the span wraps around
        # the context
        if not ignore_wrap and end_wrap - start_wrap:
            if self._abs_wrap:
                _c = self._b + abs(end_wrap - start_wrap) * l
            else:
                _c = self._b + (end_wrap - start_wrap) * l
            self._c = _c
        else:
            self._c = self._b

    @property
    def index(self):
        """Return the starting index of the context."""
        return self._index

    @property
    def a(self):
        """Return the inclusive mapped startpoint."""
        return self._a

    @property
    def b(self):
        """Return the exclusive mapped endpoint."""
        return self._b

    @property
    def c(self):
        """Return the exclusive un-mapped endpoint."""
        return self._c

    @property
    def cyclic(self):
        """Return whether the context is cyclic/circular."""
        return self._cyclic

    @property
    def context_length(self):
        """Return the length of the context."""
        return self._context_length

    def _set_as_empty(self, a: int) -> None:
        _a = self.t(a - self._index, False)
        self._a = self._b = self._c = _a
        return

    @property
    def _nwraps(self):
        return int((self._c - self._index - 1) / self._context_length)

    def bounds(self) -> tuple:
        """Return the context bounds (end exclusive)"""
        return self._index, self._context_length + self._index

    def t(self, p: int, throw_error=True) -> int:
        """Translates a position 'p' to an index within the context bounds.

        :param p:
        :type p:
        :param throw_error:
        :return:
        :rtype:
        """
        if p >= self.bounds()[1] or p < self.bounds()[0]:
            if throw_error and not self._cyclic:
                raise IndexError(
                    "Position {} outside of linear bounds {}".format(p, self.bounds())
                )
        _x = p % self._context_length
        if _x < 0:
            return self.bounds()[1] + _x
        else:
            return self.bounds()[0] + _x

    def i(self, p: int) -> int:
        """Find index of position.

        :param p:
        :return:
        """
        return p - self._a

    @staticmethod
    def _ranges_str(ranges):
        s = ",".join("[{}, {})".format(*r) for r in ranges)
        return "[" + s + "]"

    def ranges(self, ignore_wraps=False) -> List[Tuple[int, int]]:
        """Return the valid ranges for this span.

        :param ignore_wraps: if True, multiple wrappings will be ignored.
        :return: List[Tuple[int, int]]
        """
        if self._cyclic and (self._b < self._a or (self._nwraps and not ignore_wraps)):
            ranges = [(self._a, self.bounds()[1])]
            if not ignore_wraps:
                for _ in range(self._nwraps - 1):
                    ranges.append(self.bounds())
            ranges.append((self.bounds()[0], self._b))
            return ranges
        else:
            return [(self._a, self._b)]

    def slices(self):
        """Return list of slices.

        :return:
        :rtype:
        """
        return [slice(*r) for r in self.ranges()]

    def get_slice_iter(self, x: List):
        """Use the region to slice the iterable, returning another generator.

        :param x: the iterable to slice
        :return: iterable
        """
        for s in self.slices():
            yield x[s]

    def get_slice(
        self,
        x: List,
        infer_type: bool = False,
        as_type: Any = None,
        summary_func: Callable = None,
        reduce_op: Callable = operator.add,
    ) -> Any:
        """Use the region to slice the iterable, returning a iterable of the
        same type as 'x'. If as_type is provided, the iterable will be
        typecast.

        :param x: the iterable to slice
        :param as_type: type to convert the iterable into
        :param summary_func: summary function that takes in an interable. If none, uses
                            `itertools.reduce(operator.add, arr)`
        :return: Any
        """
        if summary_func is None:
            if reduce_op is not None:

                def summary_func(arr):
                    return reduce(reduce_op, arr)

            else:

                def summary_func(arr):
                    return arr

        new_arr = summary_func(self.get_slice_iter(x))
        if infer_type:
            if as_type is None:
                as_type = type(x)
        if as_type:
            return as_type(new_arr)
        return new_arr

    def reindex(self, i, strict=None, ignore_wrap=None):
        """Return a new span with positions reindexed.

        :param i: new index
        :param strict: initialize with 'strict'
        :param ignore_wrap: whether to ignore wrapping indices
        :return:
        """
        if ignore_wrap is None:
            ignore_wrap = self._ignore_wrap
        return self.new(None, None, ignore_wrap=ignore_wrap, index=i, strict=strict)

    def new(
        self,
        a: Union[int, None],
        b: Union[int, None],
        ignore_wrap=None,
        index=None,
        strict=None,
        abs_wrap=None,
    ) -> "Span":
        """Create a new span using the same context."""
        if a is None:
            a = self._a
        if b is None:
            b = self._c

        if strict is None:
            strict = self._strict

        if abs_wrap is None:
            abs_wrap = self._abs_wrap

        if ignore_wrap is None:
            ignore_wrap = self._ignore_wrap

        if index is not None:
            d = index - self._index
            a += d
            b += d
        else:
            index = self._index
        return self.__class__(
            a,
            b,
            self._context_length,
            self._cyclic,
            index=index,
            ignore_wrap=ignore_wrap,
            strict=strict,
            abs_wrap=abs_wrap,
        )

    def sub(self, a: int, b: int) -> "Span":
        """Create a sub region starting from a to b.

        :param a: starting pos
        :param b: ending pos
        :return: a new Span.
        """
        # if a == b:
        #     return self.new(a, b)
        if b is not None and a > b and not self._cyclic:
            raise ValueError(
                "Start {} cannot exceed end {} for linear spans".format(a, b)
            )

        valid_ranges = [list(x) for x in self.ranges()]
        valid_ranges[0][0] = max(a, self._a)
        valid_ranges[-1][1] = min(b + 1, self._b + 1)

        if self._nwraps > 0:
            valid_ranges += [(self._a, self._c + 1)]
        # assert len(valid_ranges) <= 2

        def in_range(pos, ranges):
            for i, r in enumerate(ranges):
                if r[0] <= pos < r[1]:
                    return True, i
            return False, None

        start_in_range, start_range = in_range(a, valid_ranges)

        if not start_in_range:
            raise IndexError(
                "Start {} must be in {}".format(a, self._ranges_str(valid_ranges))
            )

        valid_end_range = valid_ranges[start_range:]
        end_in_range, end_range = in_range(b, valid_end_range)

        if not end_in_range:
            raise IndexError(
                "End {} must be in {}".format(b, self._ranges_str(valid_end_range))
            )
        subregion = self.new(a, b)
        if len(subregion) > len(self):
            raise IndexError("Cannot make subspan.")
        return subregion

    def same_context(self, other: "Span") -> bool:
        """Return if another Span as an equivalent context."""
        return (
            other.context_length == self._context_length
            and self._cyclic == other.cyclic
        )

    def force_context(self, other: "Span") -> None:
        """Raise error if another Span has different context.

        :param other: The other span
        :return: None
        :raises: SpanError
        """
        if not self.same_context(other):
            raise SpanError("Cannot compare with different contexts")

    def overlaps_with(self, other: "Span") -> bool:
        """Returns True if other span has an overlap with this span.

        :param other: The other span
        :return: True if the two spans overlap.
        """
        self.force_context(other)
        # if other in self:
        #     return True
        # elif self in other:
        #     return True
        if (
            other.a in self
            or other.b - 1 in self
            or self._a in other
            or self._b - 1 in other
        ):
            return True
        return False

    def differences(self, other: "Span") -> Union[Tuple["Span"], Tuple["Span", "Span"]]:
        """Return a tuple of differences between this span and the other
        span."""
        self.force_context(other)
        if other.a in self and other.b not in self:
            return (self.new(self._a, other.a),)
        elif other.b in self and other.a not in self:
            return (self.new(other.b, self._b),)
        if other in self:
            return self.new(self._a, other.a), self.new(other.b, self._b)
        elif self in other:
            return (self.new(self._a, self._a),)
        else:
            return (self[:],)

    def intersection(self, other: "Span") -> "Span":
        """Return the span inersection between this span and the other span."""
        self.force_context(other)
        if other.a in self and other.b not in self:
            return self.sub(other.a, self._b)
        elif other.b in self and other.a not in self:
            return self.sub(self._a, other.b)
        if other in self:
            return other[:]
        elif self in other:
            return self[:]

    def consecutive(self, other: "Span") -> bool:
        """Returns True if other span is immediately consecutive with this
        span."""
        self.force_context(other)
        try:
            return self._b == other.t(other.a)
        except IndexError:
            return False

    def connecting_span(self, other: "Span") -> Union["Span", None]:
        """Return the span that connects the two spans. Returns None.

        :param other:
        :return:
        """
        self.force_context(other)
        if self._cyclic and self._a == other.a and self._b == other.b:
            return self.invert()[0]
        if self.consecutive(other):
            return self[self._b, self._b]
        elif self.overlaps_with(other):
            return None
        else:
            if self._b > other.a and not self._cyclic:
                return None
            return self[self._b, other.a]

    # def union(self, other):
    #     self.force_context(other)
    #     if other in self:
    #         return self[:]
    #     elif self in other:
    #         return other[:]
    #     elif other.a in self and not other.t(other.b - 1) in self:
    #         if self._a == other.b:
    #             return self.new(self._a, None)
    #         return self.new(self._a, other.b)
    #     elif other.t(other.b - 1) in self and other.a not in self:
    #         if other.a == self._b:
    #             return self.new(self._a, None)
    #         return self.new(other.a, self._b)

    # def __ge__(self, other):
    #     self.force_context(other)
    #     return self._a >= other.a

    #     def __lt__(self, other):
    #         return self._a < other.a

    #     def __gt__(self, other):
    #         return self._a > other.a

    #     def __ge__(self, other):
    #         return self._a >= other.a

    # def __invert__(self):
    #     if self._a > self._b:
    #         if self._cyclic:
    #             return self[self._b+1, self._a-1],
    #         else:
    # return

    def invert(self) -> Union[Tuple["Span", "Span"], Tuple["Span", None]]:
        """Invert the region, returning a tuple of the remaining spans from the
        context. If cyclic, a tuple (span, None) tuple is returned. If linear,
        a (span, span) is returned.

        :return: inverted regions
        :rtype: tuple
        """
        if len(self) >= self._context_length:
            return self[self._a, self._a], None
        if self._cyclic:
            return self[self._b, self._a], None
        else:
            return self[:, self._a], self[self._b, :]

    def __eq__(self, other: "Span") -> bool:
        return (
            self.same_context(other)
            and self._a == other._a
            and self._b == other._b
            and self._c == other._c
        )

    def __ne__(self, other: "Span") -> bool:
        return not (self.__eq__(other))

    @classmethod
    def _pos_in_ranges(cls, pos: int, ranges: List[Tuple[int, int]]):
        for r in ranges:
            if r[0] <= pos < r[1]:
                return True
        return False

    def contains_pos(self, pos: int) -> bool:
        """Checks if this span contains a specified position.

        :param pos: index position
        :return: bool
        """
        return self._pos_in_ranges(pos, self.ranges())

    def contains_span(self, other: "Span") -> bool:
        """Checks if this span encompasses another span.

        :param other: other span
        :return: bool
        """
        if not self.same_context(other):
            return False
        if other.a == other.b == self._a:
            # special case where starting indices and ending indices are the same
            return True
        else:
            if not self.contains_pos(other.a):
                return False
            elif not self.contains_pos(other.t(other.b - 1)):
                return False
            elif not len(other) <= len(self):
                return False
            return True

    def spans_origin(self) -> bool:
        if self._nwraps and self._cyclic:
            return True
        return self._b < self._a and self._cyclic

    def __contains__(self, other: Union["Span", int]):
        if isinstance(other, int):
            return self.contains_pos(other)
        elif issubclass(type(other), Span):
            return self.contains_span(other)

    def __len__(self) -> int:
        return sum([r[1] - r[0] for r in self.ranges()])

    def __iter__(self) -> Generator[int, None, None]:
        for i in chain(*[range(*x) for x in self.ranges()]):
            yield i

    def __invert__(self):
        return self.invert()

    def _check_index_pos(self, val: int, inclusive=True) -> None:
        if val is not None and not self.cyclic:

            if inclusive and (val >= len(self) or val < -len(self)):
                raise IndexError(
                    "Index '{}' outside of linear span with length of {}".format(
                        val, len(self)
                    )
                )
            elif not inclusive and (val > len(self) or val < -len(self)):
                raise IndexError(
                    "Exclusive index '{}' outside of linear span with length of {}".format(
                        val, len(self)
                    )
                )

    def __getitem__(self, val):
        if isinstance(val, int):
            self._check_index_pos(val)
            if val < 0:
                return self.t(val + self._c - self._index)
            else:
                return self.t(val + self._a - self._index)
        elif issubclass(type(val), slice):
            self._check_index_pos(val.start)
            self._check_index_pos(val.stop, False)

            if val.step == -1:
                return self[val.start : val.stop].invert()
            elif val.step is not None and val.step != 1:
                raise ValueError(
                    "{} slicing does not support step {}.".format(
                        self.__class__.__name__, val.step
                    )
                )

            if val.start is None:
                i = self._a
            elif val.start < 0:
                i = self._c + val.start
            else:
                i = self._a + val.start

            if val.stop is None:
                j = self._c
            elif val.stop < 0:
                j = self._c + val.stop
            else:
                j = self._a + val.stop

            return self.new(i, j)
        elif isinstance(val, tuple):
            if len(val) > 2:
                raise ValueError(
                    "{} -- copying only supports (start, stop)".format(val)
                )
            val = list(val)
            for i in [0, 1]:
                if val[i] == slice(None, None, None):
                    val[i] = None
            if val == (None, None):
                val = self.bounds()
            elif val[0] is None:
                val = (self.bounds()[0], val[1])
            elif val[1] is None:
                val = (val[0], self.bounds()[1])
            return self.new(*val)
        else:
            raise ValueError("indexing does not support {}".format(type(val)))

    def __repr__(self):
        return (
            "<{cls} {a} {b} ({c}) context_len={context_len} len={length} "
            "cyclic={cyclic} index={index}, nwraps={n}>".format(
                cls=self.__class__.__name__,
                a=self._a,
                b=self._b,
                c=self._c,
                context_len=self._context_length,
                cyclic=self._cyclic,
                index=self._index,
                length=len(self),
                n=self._nwraps,
            )
        )

    def __str__(self):
        return self.__repr__()


class EmptySpan(Span):
    def ranges(self, *args) -> List[Tuple[int, int]]:
        return [(self.bounds()[0], self.bounds()[0])]


class Direction:
    FORWARD = 1
    REVERSE = -1
    BOTH = 0


class Region(Span):
    """A direction span."""

    __slots__ = ["name", "id", "direction"]

    FORWARD = Direction.FORWARD
    REVERSE = Direction.REVERSE
    BOTH = Direction.BOTH

    def __init__(
        self,
        start: int,
        end: int,
        length: int,
        cyclic: bool = False,
        index: int = 0,
        direction: int = FORWARD,
        name: Union[None, str] = None,
        region_id: Union[None, str] = None,
        ignore_wrap: bool = False,
        abs_wrap: bool = True,
        strict: bool = None,
    ):
        """Initialize a new Region.

        :param start: start position
        :param end: end position (exclusive)
        :param length: length of context
        :param cyclic: topology of context
        :param index: starting index of context
        :param direction: direction of the region
        :param name: name of the region
        :param region_id: id of the region
        :param strict: if True, positions outside of context bounds are disallowed.
        :type strict: bool
        :param ignore_wrap: if True (default False), initialization indicies that wrap
                            around multiple times will
                            simply be mapped directly to the context (no wrapping used).
        :type ignore_wrap: bool
        :param abs_wrap:    if True, the abs difference between start and end wrappings
                            are used. Starting wraps that are greater than ending wraps
                            are valid. If False and the starting wrap is greater than
                            the ending wrap, an IndexError is thrown.
        """
        self.name = name
        self.id = region_id
        assert direction in [self.FORWARD, self.REVERSE, self.BOTH]
        self.direction = direction
        super().__init__(
            start,
            end,
            length,
            cyclic=cyclic,
            index=index,
            ignore_wrap=ignore_wrap,
            strict=strict,
            abs_wrap=abs_wrap,
        )

    def flip(self) -> "Region":
        """Flip the indices of the region."""
        flipped = self.new(self.context_length - self.b, self.context_length - self.a)
        flipped.direction *= -1
        return flipped

    @property
    def start(self):
        if self.direction == self.REVERSE:
            return self.b
        else:
            return self.a

    @property
    def end(self):
        if self.direction == self.REVERSE:
            return self.a
        else:
            return self.b

    def new(self, a: Union[None, int], b: Union[None, int]):
        _new = super().new(a, b)
        _new.direction = self.direction
        return _new

    def __eq__(self, other: "Region") -> bool:
        return super().__eq__(other) and self.direction == other.direction

    def __str__(self):
        return (
            "<{cls} {a} {b} ({c}) context_len={context_len} len={length} "
            "cyclic={cyclic} direction={direction} index={index}, nwraps={n}>".format(
                cls=self.__class__.__name__,
                a=self._a,
                b=self._b,
                c=self._c,
                context_len=self._context_length,
                cyclic=self._cyclic,
                index=self._index,
                length=len(self),
                n=self._nwraps,
                direction=self.direction,
            )
        )



"""sequence_generator."""
import random
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from uuid import uuid4

from Bio.SeqRecord import SeqRecord
from pyblast.utils import make_circular
from pyblast.utils import make_linear

from dasi.utils import biopython


def iter_fake_random_record(
    n_seqs: int, size_int: Tuple[int, int], cyclic: bool
) -> List[SeqRecord]:
    for i in range(n_seqs):
        length = random.randint(*size_int)

        name = "<random record {}".format(str(uuid4()))
        rec = biopython.random_record(length, name=name, auto_annotate=True)
        rec.id = rec.name
        biopython.randomly_annotate(rec, (100, 1000))

        if cyclic:
            make_circular([rec])
        else:
            make_linear([rec])
        yield rec


def fake_library(
    n_cyclic_seqs: int,
    n_linear_seqs: int,
    n_primers: int,
    cyclic_size_int: Tuple[int, int] = (3000, 10000),
    linear_size_int: Tuple[int, int] = (100, 4000),
    primer_size_int: Tuple[int, int] = (15, 60),
    return_all: bool = False,
) -> Union[List[SeqRecord], Dict[str, List[SeqRecord]]]:
    cyclic_seqs = list(
        iter_fake_random_record(n_cyclic_seqs, size_int=cyclic_size_int, cyclic=True)
    )
    linear_seqs = list(
        iter_fake_random_record(n_linear_seqs, size_int=linear_size_int, cyclic=False)
    )
    primer_seqs = list(
        iter_fake_random_record(n_primers, size_int=primer_size_int, cyclic=False)
    )
    if return_all:
        return cyclic_seqs + linear_seqs + primer_seqs
    else:
        return {"cyclic": cyclic_seqs, "linear": linear_seqs, "short": primer_seqs}


def _add_shared_sequence(
    sequences: List[SeqRecord], length: int, name="shared sequence"
):
    new_designs = []
    shared_chunk = biopython.random_record(length, name=name)
    for design in sequences:
        i = random.randint(0, len(design))
        new_designs.append(design[:i] + shared_chunk + design[i:])
    return new_designs


# TODO: replace with with MC method
def fake_designs(
    n_designs: int,
    circular: bool,
    n_cyclic_seqs: int,
    n_linear_seqs: int,
    n_primers: int,
    n_primers_from_templates: int,
    design_sequence_similarity_length: int = 0,
    cyclic_size_int: Tuple[int, int] = (3000, 10000),
    linear_size_int: Tuple[int, int] = (100, 4000),
    primer_size_int: Tuple[int, int] = (15, 60),
    plasmid_size_interval: Tuple[int, int] = (5000, 10000),
    chunk_size_interval: Tuple[int, int] = (100, 3000),
    random_chunk_prob_int: Tuple[float, float] = (0, 0.5),
    random_chunk_size_int: Tuple[int, int] = (100, 1000),
):
    library_dict = fake_library(
        n_cyclic_seqs=n_cyclic_seqs,
        n_linear_seqs=n_linear_seqs,
        n_primers=n_primers,
        cyclic_size_int=cyclic_size_int,
        linear_size_int=linear_size_int,
        primer_size_int=primer_size_int,
    )
    linear_seqs = library_dict["linear"]
    cyclic_seqs = library_dict["cyclic"]
    templates = cyclic_seqs + linear_seqs
    short_seqs = library_dict["short"]

    for i in range(n_primers_from_templates):
        primer = biopython.random_record_from_library(
            templates,
            circular=False,
            size_interval=(15, 100),
            max_chunks=1,
            chunk_size_interval=(15, 60),
            random_chunk_prob_int=(0, 0),
            random_chunk_size_int=(0, 0),
        )
        short_seqs.append(primer)

    # generate designs from templates or random sequence
    designs = []
    for i in range(n_designs):
        rec = biopython.random_record_from_library(
            templates,
            circular=circular,
            size_interval=plasmid_size_interval,
            chunk_size_interval=chunk_size_interval,
            random_chunk_prob_int=random_chunk_prob_int,
            random_chunk_size_int=random_chunk_size_int,
        )
        designs.append(rec)

    if design_sequence_similarity_length:
        designs = _add_shared_sequence(designs, design_sequence_similarity_length)

    if circular:
        make_circular(designs)
    else:
        make_linear(designs)

    return {
        "design": designs,
        "cyclic": cyclic_seqs,
        "linear": linear_seqs,
        "short": short_seqs,
    }



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['dasi',
 'dasi.command_line',
 'dasi.cost',
 'dasi.design',
 'dasi.models',
 'dasi.schemas',
 'dasi.utils',
 'dasi.utils.biopython',
 'dasi.utils.networkx']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.74,<2.0',
 'fire>=0.1,<0.2',
 'frozendict>=1.2,<2.0',
 'jsonschema>=3.1,<4.0',
 'loggable-jdv>=0.1.5,<0.2.0',
 'matplotlib>=3.1,<4.0',
 'more-itertools>=8.0,<9.0',
 'msgpack-numpy>=0.4.4,<0.5.0',
 'msgpack>=0.6.1,<0.7.0',
 'nest_asyncio>=1.0,<2.0',
 'networkx>=2.3,<3.0',
 'numpy>=1.17,<2.0',
 'pandas>=0.25.1,<0.26.0',
 'primer3plus>=1.0.5,<2.0.0',
 'pyblastbio>=0.6.3,<0.7.0',
 'seaborn>=0.9.0,<0.10.0',
 'sortedcontainers>=2.1,<3.0',
 'sympy>=1.4,<2.0',
 'tqdm>=4.32,<5.0',
 'uvloop>=0.12.2,<0.13.0']

entry_points = \
{'console_scripts': ['dasi = dasi:command_line.main']}

setup_kwargs = {
    'name': 'dasi',
    'version': '0.0.13',
    'description': 'Automated DNA assembly planner for Python',
    'long_description': "# DASi DNA Design\n\n[![PyPI version](https://badge.fury.io/py/dasi.svg)](https://badge.fury.io/py/dasi)\n\n**DASi** is an automatic DNA cloning plan designer aimed for operating on small budgets\nby focusing on material re-use.\n\nThe software converts a nucleotide sequence, or a library of sequences, to an executable\n molecular assembly plan while optimizing material cost, assembly efficiency, and assembly time.\n\nThe key design paradigm for DASi is that *no molecular biology expertise* is required to use DASi. Complete novices should be able to use the software to design and construct new genetic sequences. This also enables automated software programs to automatically design and construct new genetic sequences.\n\nThe software goals are reminiscent of j5 or Teselegen but focused on:\n1. A dead-simple API usable by lab novices, experts or automated software programs.\n1. Utilizing information about current laboratory inventory in its optimization\nalgorithm to minimize costs and turn-around time\n\n### Status\n\nDASi is currently under development funded by the DARPA Synergistic Discovery and Design program. DASi is currently being used to connect automatically generate DNA designs to automated biological fabrication facilities (e.g. University of Washington Biofab).\n\n### Usage\n\nDASi completely automates the cloning design work, finding approximately optimal solutions for cloning steps, preferentially using existing plasmids, linear DNA fragments, and primers to design semi-optimal cloning steps and designs.\n\nThe following command designs the cloning steps for a library of designs. The user only needs to specify the sequences they wish to construct and currently available primers and DNA templates as *.genbank* or *.fasta* files. DASi handles all design aspects. *No molecular biology expertise is required to use DASi.*\n\n```bash\ndasi library_design --designs mydesigns/*.gb --fragments fragments/*.gb --primers primers.fasta --templates plasmids/*.gb --cost_model cost.b --out results\n```\n\n#### Customization\n\nDASi optimization parameters are completely customizable. The following are examples of parameters and aspects of DASi that are customizable:\n\n* primer synthesis costs\n* primer design parameters\n* synthetic fragment costs\n* vendor-specific synthetic fragment complexity\n* sequence dependent plasmid assembly efficiencies\n* optimizing over efficiency vs material costs\n* etc.\n\n### Planned Features\n\n* Golden-gate support\n* heirarchical assembly\n* library support (with bayesian search to optimize shared parts)\n* front-end\n* connection to fabrication facility\n\n### DASi optimization problem\n\nBriefly, DASi approximates a solution the following optimization problem: \n\n```Given a set of 'goal' double-stranded sequences, a set of available single-stranded and double-strand sequences, and a set of actions that can create new sequences, find the optimal set of operations that produces the 'goal' sequences.```\n\nFormalization of this optimization problem is coming soon.\n",
    'author': 'Justin Vrana',
    'author_email': 'justin.vrana@gmail.com',
    'url': 'https://github.com/jvrana/dasi-dna-design',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from heapq import heappop
from heapq import heappush
from itertools import count
from typing import List

import networkx as nx
import numpy as np
from more_itertools import pairwise
from sympy import lambdify
from sympy import sympify

from .exceptions import NetworkxUtilsException
from .utils import sort_cycle


def _weight_function(v, u, e, k):
    if e is None:
        return None
    return e[k]


def sympy_multisource_dijkstras(
    g, sources, f, accumulators=None, init=None, target=None, cutoff=None
):
    if sources is None or not len(sources):
        raise ValueError("sources must not be empty")
    if target in sources:
        return (0, [target])
    paths = {source: [source] for source in sources}  # dictionary of paths
    dist = _multisource_dijkstra(
        g,
        sources,
        f,
        target=target,
        accumulators=accumulators,
        init=init,
        paths=paths,
        cutoff=cutoff,
    )
    if target is None:
        return (dist, paths)
    try:
        return (dist[target], paths[target])
    except KeyError:
        raise nx.NetworkXNoPath("No path to {}.".format(target))


def sympy_dijkstras(
    g, source, f, target=None, accumulators=None, init=None, cutoff=None
):
    """Computes the shortest path distance and path for a graph using an
    arbitrary function.

    :param g:
    :param source:
    :param f:
    :param target:
    :param accumulators:
    :param init:
    :param cutoff:
    :return:
    """
    dist, path = sympy_multisource_dijkstras(
        g,
        [source],
        f,
        target=target,
        accumulators=accumulators,
        init=init,
        cutoff=cutoff,
    )
    return dist, path


def _multisource_dijkstra(
    g,
    sources,
    f,
    target=None,
    accumulators=None,
    init=None,
    cutoff=None,
    paths=None,
    pred=None,
):
    accumulators = accumulators or {}
    init = init or {}

    # successor dictionary
    g_succ = g._succ if g.is_directed() else g._adj

    # push/pop methods to use
    push = heappush
    pop = heappop

    # node to shortest distance, breakdown
    dist_parts = {}

    # node to shortest distance
    dist = {}

    # node to shortest distance seen
    seen = {}
    c = count()
    fringe = []

    # lambda function
    func = sympify(f)
    symbols = list([s.name for s in func.free_symbols])
    func = lambdify(symbols, func)

    # initial/default values for each symbol
    for sym in symbols:
        if accumulators.get(sym, "sum") == "sum":
            init.setdefault(sym, 0.0)
        elif accumulators.get(sym, "product"):
            init.setdefault(sym, 1.0)
        else:
            raise NetworkxUtilsException("Accumulator '{}' not recognized".format(sym))
    init = np.array([init[x] for x in symbols])

    # accumulator functions to each for each symbol
    # if 'product', return the accumulated product
    # else return the accumulated sum (default)
    accu_f = []
    for sym in symbols:
        if accumulators.get(sym, "sum") == "sum":
            accu_f.append(lambda x: np.sum(x))
        elif accumulators[sym] == "product":
            accu_f.append(lambda x: np.prod(x))

    # push the initial values for the sources
    for source in sources:
        if source not in g:
            raise nx.NodeNotFound("Source {} is not in G".format(source))
        seen[source] = func(*init)
        push(fringe, (0, next(c), source, init))

    # modified dijkstra's
    while fringe:
        (_, _, v, d) = pop(fringe)
        # d np.array of values
        if v in dist_parts:
            continue  # already searched this node
        dist_parts[v] = d
        dist[v] = func(*d)
        if v == target:
            break
        for u, e in g_succ[v].items():
            # vu cost breakdown for  each symbol
            costs = np.array([_weight_function(v, u, e, sym) for sym in symbols])

            # vu_dist break down using accumulating function
            x = np.stack([dist_parts[v], costs])
            vu_dist_parts = []
            for i, _x in enumerate(x.T):
                vu_dist_parts.append(accu_f[i](_x))
            vu_dist_parts = np.array(vu_dist_parts)
            vu_dist = func(*vu_dist_parts)

            if cutoff is not None:
                if vu_dist > cutoff:
                    continue
            if u in dist_parts:
                if vu_dist < dist[u]:
                    raise ValueError("Contradictory paths found:", "negative weights?")
            elif u not in seen or vu_dist < seen[u]:
                seen[u] = func(*vu_dist_parts)
                push(fringe, (vu_dist, next(c), u, vu_dist_parts))
                if paths is not None:
                    paths[u] = paths[v] + [u]
                if pred is not None:
                    pred[u] = v
            elif vu_dist == seen[u]:
                if pred is not None:
                    pred[u].append(v)
                    fringe = []
    return dist


def multipoint_shortest_path(
    graph: nx.DiGraph,
    nodes: List[str],
    weight_key: str,
    cyclic=False,
    cyclic_sort_key=None,
):
    """Return shortest path through nodes. If cyclic, will return the cycle
    sorted with the 'lowest' node at index 0. Self cycles are not supported.

    :param graph: the graph
    :param nodes: list of nodes to find path
    :param weight_key: weight key
    :param cyclic: whether the path is cyclic
    :param cyclic_sort_key: the key function to use to sort the cycle (if cyclic)
    :return:
    """
    if cyclic_sort_key and not cyclic:
        raise ValueError("cyclic_sort_key was provided but 'cyclic' was False.")
    full_path = []
    if cyclic:
        nodes = nodes + nodes[:1]
    for n1, n2 in pairwise(nodes):
        path = nx.shortest_path(graph, n1, n2, weight=weight_key)
        full_path += path[:-1]
    if not cyclic:
        full_path.append(nodes[-1])
    if cyclic:
        return sort_cycle(full_path, cyclic_sort_key)
    else:
        return full_path


def sympy_multipoint_shortest_path(
    graph: nx.DiGraph,
    nodes: List[str],
    f: str,
    accumulators: dict,
    init=None,
    cutoff=None,
    cyclic=False,
    cyclic_sort_key=None,
):
    if cyclic_sort_key and not cyclic:
        raise ValueError("cyclic_sort_key was provided but 'cyclic' was False.")
    full_path = []
    full_path_length = 0.0
    if cyclic:
        nodes = nodes + nodes[:1]
    for n1, n2 in pairwise(nodes):
        path_length, path = sympy_dijkstras(
            graph,
            f=f,
            source=n1,
            target=n2,
            accumulators=accumulators,
            init=init,
            cutoff=cutoff,
        )
        full_path_length += path_length
        full_path += path[:-1]
    if not cyclic:
        full_path.append(nodes[-1])
    if cyclic:
        return full_path_length, sort_cycle(full_path, cyclic_sort_key)
    else:
        return full_path_length, full_path
